#ifndef _GI_PANGO_PANGO_SETUP_DEF_HPP_
#define _GI_PANGO_PANGO_SETUP_DEF_HPP_

/* some header parts are only conditionally considered */
#define PANGO_ENABLE_BACKEND
#define PANGO_ENABLE_ENGINE

#endif
