#ifndef _GI_GST_GL_EGL_SETUP_PRE_DEF_HPP_
#define _GI_GST_GL_EGL_SETUP_PRE_DEF_HPP_

// missing G_BEGIN_DECLS / G_END_DECLS
// so wrap those around and pre-include
G_BEGIN_DECLS
#include <gst/gl/egl/gstegl.h>
G_END_DECLS

#endif // _GI_GST_GL_EGL_SETUP_PRE_DEF_HPP_
